using System.Collections;
using System.Windows.Forms;
using System.Drawing;

namespace Solitare {
	
   /// <summary>
   /// Allows a Card Game to define the state of a move for undo purposes
   /// </summary>
   public enum MoveStatus : int {
      /// <summary>
      /// Click was valid, but does not end the move.
      /// </summary>
      ValidClick, 
      /// <summary>
      /// Click was valid, and does end the move
      /// </summary>
      ValidMoveEnd, 
      /// <summary>
      /// Click was not valid, and terminates the move
      /// </summary>
      Invalid,
      /// <summary>
      /// Action was not applicable to this game.
      /// </summary>
      MoveNotApplicable
   }
	
   /// <summary>
   /// When searching for valid wins, how should the game react, especially to a Victory condition.
   /// </summary>
   public enum MoveSearchType {
      /// <summary>
      /// Stop if a valid win has been calculated.
      /// </summary>
      SingleWin
   }
   
   public struct CardMoveEventArgs {
      public Point OriginalLocation;
      public Point NewLocation;
   }

   public struct ValidMoveEventArgs {
      public ArrayList MoveList;
   }
	
   public abstract class IRules {
      /// <summary>
      /// Used to build the initial Deck. Returns a reference to the public Card Stack which
      /// will be shuffled and added to the Board.
      /// </summary>
      /// <param name="Cards">An out array of Cards which will be added the the board, and shuffled into the deck.</param>
      /// <param name="Deck">A reference to the public Deck object.</param>
      public abstract void BuildDeck(out PlayingCard[] Cards, out PlayingStack Deck);

      /// <summary>
      /// Draws the board, laying out where the deck goes, where the stacks go, and initial cards.
      /// </summary>
      /// <param name="Board">The Board Control that the stacks will be added to.</param>
      /// <param name="PlayingStacks">The stacks that need event (click) monitoring.</param>
      public abstract void LayoutBoard(Panel Board, out PlayingStack[] PlayingStacks);
      
      /// <summary>
      /// Determines how best to handle a dropped card. (LMB released.)
      /// </summary>
      /// <param name="Card">The Card dropped.</param>
      /// <param name="OntoStack">The stack onto which the card was dropped.</param>
      public abstract MoveStatus HandleDroppedCard(PlayingCard Card, PlayingStack OntoStack);

      /// <summary>
      /// Denotes that a card has been double clicked.
      /// </summary>
      /// <param name="Card">The Card that was clicked on.</param>
      public abstract MoveStatus HandleClickedCard(PlayingCard Card);

      /// <summary>
      /// Denotes that a card has been clicked.
      /// </summary>
      /// <param name="Card">The Card that was clicked on.</param>
      public abstract MoveStatus HandleDoubleClickedCard(PlayingCard Card);

      /// <summary>
      /// Denotes that a empty stack has been clicked.
      /// </summary>
      /// <param name="Stack">The stack that received the click event.</param>
      public abstract MoveStatus HandleClickedStack(PlayingStack Stack);

      /// <summary>
      /// Denotes that a empty stack has been double clicked.
      /// </summary>
      /// <param name="Stack">The stack that received the click event.</param>
      public abstract MoveStatus HandleDoubleClickedStack(PlayingStack Stack);

      /// <summary>
      /// Used on mouse down to determine if the selected card can move.
      /// </summary>
      /// <param name="Card">The Card to check</param>
      /// <returns>True if it can be moved (dragged), false otherwise.</returns>
      public abstract bool CanCardMove(PlayingCard Card);

      /// <summary>
      /// Used to check for Victory Conditions.
      /// </summary>
      /// <returns>Returns true if Victory Conditions have all been met.</returns>
      public abstract bool CheckForVictory();

      public object GameSpecificInfo = new object();
      
      public abstract ArrayList GetValidMoves();

      public MoveSearchType MoveSearchType = MoveSearchType.SingleWin;

      protected int _maxMoves = -1;
      public int MaxMoves {
         get { return _maxMoves; }
      }


      /// <summary>
      /// Current score of the game. 
      /// </summary>
      public abstract int Score { get; }

      public event CardMoveEventHandler CardMoveEvent;

      protected void OnCardMove(PlayingCard card, System.Drawing.Point newLoc) {
         CardMoveEventArgs e = new CardMoveEventArgs();
         
         e.OriginalLocation = card.Location;
         e.NewLocation = newLoc;

         if (CardMoveEvent != null) {
            CardMoveEvent(card, e);
         }
      }

      public event ValidMoveEventHandler ValidMoveEvent;

      protected void OnValidMove() {
         ArrayList al = GetValidMoves();

         ValidMoveEventArgs vme = new ValidMoveEventArgs();
         vme.MoveList = al;

         if (ValidMoveEvent != null) {
            ValidMoveEvent(this, vme);
         }
      }

      public event System.EventHandler StoreUndoMoveEvent;

      protected void OnStoreUndoMove() {
         StoreUndoMoveEvent(this, System.EventArgs.Empty);
      }

   }
   public delegate void CardMoveEventHandler(PlayingCard Card, CardMoveEventArgs e);
   public delegate void ValidMoveEventHandler(IRules sender, ValidMoveEventArgs e);
}