using System;
using System.Drawing;
using System.Windows.Forms;
using System.Collections;

namespace Solitare {

	public enum StackTypes : int {
		Deck = 0,
      Discard = 1,
		Foundation = 2,
		Tableu = 3,
      Other = 4
	}
	
	public class PlayingStack : Panel {
      internal static Size stackSize = 
         new System.Drawing.Size(
         PlayingCard.cardSize.Width + 2,
         PlayingCard.cardSize.Height + 2);

      private int _stackId;
      public int StackId {
         get { return _stackId; }
      }
      
		private StackTypes _type;
		public StackTypes Type {
			get { return _type; }
			set { _type = value; }
		}
		
		private bool _drawTiled;
      internal bool DrawTiled {
         get { return _drawTiled; }
         set { _drawTiled = value; }
      }

      private int _hashcode = 0;
      public int Hashcode {
         get { return _hashcode; }
         set { _hashcode = value; }
      }
      
      public new Point Location {
         set { 
            Point p = value;
            p.X -= 1;
            p.Y -= 1;
            base.Location = p;
         }

         get { 
            Point p = base.Location;
            p.X += 1;
            p.Y += 1;
            return p;
         }
      }

      public Rectangle Rect {
         get { return new Rectangle(this.Location.X, this.Location.Y, this.Width, this.Height); }
      }

      public bool IntersectsWith(PlayingCard Card) {
         PlayingCard top = LookTopCard();
         if (top == null) {
            return (Rect.IntersectsWith(Card.Rect)); 
         } else {
            return (top.Rect.IntersectsWith(Card.Rect)); 
         }
      }

      public long IntersectingArea(PlayingCard Card) {
         if (!IntersectsWith(Card) ) {
            return -1;
         }
         PlayingCard top = LookTopCard();
         Rectangle test;
         if (top == null) {
            test = this.Rect;
         } else {
            test = top.Rect;
         }

         Rectangle i = Rectangle.Intersect(this.Rect, Card.Rect);
         return i.Width * i.Height;
      }

      private CardCollection _cards = new CardCollection();
      public CardCollection Cards {
         get { return _cards; }
      }

      internal int CardCount {
         get { return _cards.Count; }
      }

      private int calculateHash() {
         _hashcode = 0;
         for (int i = 0; i < _cards.Count; i++) {
            PlayingCard card = _cards[i];

            _hashcode += ((i + 1) * 2 * 52) * Convert.ToInt32(card.Number) * (Convert.ToInt32(card.Suite) + 1);
         }
         return _hashcode;
      }
      
      internal PlayingCard GetTopCard() {
         if (_cards.Count > 0) {
            OnRemoveCard(_cards.Peek());
            PlayingCard top = _cards.Pop();
            if (_cards.Count > 0) {
               PlayingCard newTop = _cards.Peek();

               newTop.UnderCard = null;
            }
            top.OverCard = null;

            return top;
         } else 
            return null;
      }

      internal PlayingCard LookTopCard() {
         if (_cards.Count > 0) {
            return _cards.Peek();
         } else {
            return null;
         }
      }

      // This logic may seem a little wacky, but it's too pull
      // a card out from the middle of the deck, and maintain the links (I hope it works)
      internal bool RemoveCard(PlayingCard Card) {
         PlayingCard overCard = Card.OverCard;
         PlayingCard underCard = Card.UnderCard;

         if (overCard != null) {
            if (underCard != null) {
               overCard.UnderCard = underCard;
            } else {
               overCard.UnderCard= null;
            }
         }

         if (underCard != null) {
            if (overCard != null) {
               underCard.OverCard = underCard;
            } else {
               underCard.OverCard = null;
            }
         }

         Card.OverCard = null;
         Card.UnderCard = null;

         calculateHash();

         OnRemoveCard(Card);

         return _cards.Remove(Card);
      }
      
      internal bool RemoveTopCard(PlayingCard Card) {
         if (_cards.Count > 0 && _cards.Peek() == Card) {
            OnRemoveCard(Card);
            
            PlayingCard top = _cards.Pop();
            
            if (_cards.Count > 0) {
               PlayingCard newTop = _cards.Peek();

               newTop.UnderCard = null;
            }

            top.OverCard = null;

            calculateHash();

            return true;
         } else {
            return false;
         }
      }

      internal void AddTopCard(PlayingCard Card) {
         if (Card.Stack != null) {
            Card.Stack.RemoveTopCard(Card);
         }
         
         if (_cards.Count > 0) {
         PlayingCard currentTop = _cards.Peek();
            currentTop.UnderCard = Card;
            Card.OverCard = currentTop;
         }
         
         _cards.Push(Card);
         Card.BringToFront();
         if (_drawTiled) {
            int offsetNew = 0;
            for (int i = 0; i < _cards.Count - 1; i++) {
               if (_cards[i].IsFaceShowing) {
                  offsetNew += 30;
               } else {
                  offsetNew += 10;
               }
            }
            int offset = (_cards.Count -1) * 30;
            Card.Location = new Point(this.Location.X, this.Location.Y + offsetNew);
         } else {
            Card.Location = this.Location;
         }

         Card.Stack = this;

         calculateHash();

         OnAddCard(Card);
      }

      internal PlayingStack(int StackId) : base() {
         BackColor = Color.Transparent;
         Size = stackSize;
         this._stackId = StackId;
      }

      protected override void OnPaint(PaintEventArgs e) {
         Graphics g = e.Graphics;
         this.ForeColor = Color.Black;
         Pen p = new Pen(Color.Red);
         g.DrawLine(p, new Point(0,0), new Point(0, this.Height - 1));
         g.DrawLine(p, new Point(0,0), new Point(this.Width - 1, 0));

         g.DrawLine(p, new Point(0, this.Height - 1), new Point(this.Width - 1, this.Height - 1));
         g.DrawLine(p, new Point(this.Width - 1, 0), new Point(this.Width - 1, this.Height - 1));
      }

      public static event EventHandler RemoveCardEvent;
      public static event EventHandler AddCardEvent;

      public static void OnRemoveCard(PlayingCard Card) {
         RemoveCardEvent(Card, EventArgs.Empty);
      }

      public static void OnAddCard(PlayingCard Card) {
         AddCardEvent(Card, EventArgs.Empty);
      }

   }
}