/* For undo functionality, simple copy each card move that comprises a move.
 * Single cards will move indvidual, and stacks will move as stacks.
 * Cards turned over will be tracked, as well as pulled from the deck.
 * Score will be tracked as well.
 */

/* Redoing undo calculations.
 * - Still store them as discreet moves, but...
 * 
 * 1) Calculate full hash for playing area.
 * 2) When MoveCards or FlipCard methods are executed, they throw an event to the MoveCard listener.
 * 3) Each event is added to the current move list.
 * 4) Recheck the hash. If the hash is different (meaning the move wasn't undone), store the undo, and
 *    the new hash.
 *    NOTE: This calculation might be better handled if based off of the ValidMoveEnd
 * 
 * NOTE: !!!! While tracking moves, any information that may change internal depending on the state of the
 * cards must be stored in a property call XXXXX that can be restored after the undo.
 */


using System;
using System.Collections;
using System.Diagnostics;

using System.IO;
using System.Runtime.Serialization.Formatters.Soap;

namespace Solitare {

   public class StoreUndoMoves {
   
      Stack _undoMoves;

      Stack _validMoves;

      int currentHash = -1;

      bool inUndo = false;

      private String _currentHash;
      public String CurrentHash {
         get { return _currentHash; }
         set { _currentHash = value; }
      }

      public StoreUndoMoves() {
         PlayingCard.CardFlipEvent += new EventHandler(OnCardFlip);
         PlayingStack.AddCardEvent += new EventHandler(OnAddCard);
         PlayingStack.RemoveCardEvent += new EventHandler(OnRemoveCard);

         _undoMoves = new Stack();
         _validMoves = new Stack();
      }

      private void OnCardFlip(Object sender, EventArgs e) {
         if (!inUndo) {
            PlayingCard card = (PlayingCard)sender;
            _undoMoves.Push(new CardFlipAction(card));
         }
      }

      private void OnRemoveCard(Object sender, EventArgs e) {
         if (!inUndo) {
            PlayingCard card = (PlayingCard)sender;
            _undoMoves.Push(new CardRemoveAction(card));
         }
      }
      
      private void OnAddCard(Object sender, EventArgs e) {
         if (!inUndo) {
            PlayingCard card = (PlayingCard)sender;
            _undoMoves.Push(new CardAddAction(card));
         }
      }

      String currentGsi;
      public void StartTrackingMove(int Hash, IRules CurrentGame) {
         currentGsi = SerializeGameSpecificInfo(CurrentGame.GameSpecificInfo);
         _undoMoves = new Stack();
         currentHash = Hash;
      }

      public void EndTrackingMove(int Hash, IRules CurrentGame) {
         String gsi = SerializeGameSpecificInfo(CurrentGame.GameSpecificInfo);

         if (gsi != currentGsi) {
            GameSpecificChangedAction gsc = new GameSpecificChangedAction(CurrentGame, currentGsi);
            _undoMoves.Push(gsc);
         }

         if (currentHash != Hash) {
            _validMoves.Push(_undoMoves);
         }
      }

      private String SerializeGameSpecificInfo(Object GameSpecificInfo) {
         SoapFormatter sf = new SoapFormatter();
         MemoryStream ms = new MemoryStream();
         sf.Serialize(ms, GameSpecificInfo);
         
         String s = System.Text.ASCIIEncoding.ASCII.GetString(ms.ToArray());

         return s;
      }

      private Object DeserializeGameSpecificInfo(String GameSpecificInfo) {
         SoapFormatter sf = new SoapFormatter();
         MemoryStream ms = new MemoryStream(System.Text.ASCIIEncoding.ASCII.GetBytes(GameSpecificInfo));

         return sf.Deserialize(ms);
      }



      public void UndoLastMove() {
         inUndo = true;
         if (_validMoves.Count > 0) {
            Stack moveList = (Stack)_validMoves.Pop();
         
            while (moveList.Count > 0) {
               UndoMoveAction uma = (UndoMoveAction)moveList.Pop();

               if (uma is CardFlipAction) {
                  uma.Card.IsFaceShowing = ((CardFlipAction)uma).FaceVisibleAtStart;
               } else if (uma is CardRemoveAction) {
                  // TODO This will *NOT* work for removed cards...
                  ((CardRemoveAction)uma).RemovedFromStack.AddTopCard(uma.Card);
               } else if (uma is CardAddAction) {
                  ((CardAddAction)uma).AddedToStack.RemoveCard(uma.Card);
               } else if (uma is GameSpecificChangedAction) {
                  String s = ((GameSpecificChangedAction)uma).GameSpecificInfo;
                  Object o = DeserializeGameSpecificInfo(s);

                  ((GameSpecificChangedAction)uma).CurrentGame.GameSpecificInfo = o;
               } else {
                  Trace.WriteLine("Invalid Undo Type");
               }
            }
         }
         inUndo = false;
      }

      public void ResetUndoMoves() {
         _undoMoves = new Stack();
      }
   }
   
   public class ValidMove {
      // This will be a list, in order, of the cards & stacks clicked on.
      // This information will define the move
      private ArrayList _cardList;
      public ArrayList CardList {
         get { return _cardList; }
         set { _cardList = value; }
      }

      private int _scoreChange;
      public int ScoreChange {
         get { return _scoreChange; }
         set { _scoreChange = value; }
      }


      // This will define, in order, the changes made. What card(s) moved to
      // where, what cards were flipped as a result, what cards were drawn from
      // the desk to replenish empty stacks.
      private ArrayList _changeList;
      public ArrayList ChangeList {
         get { return _changeList; }
         set { _changeList = value; }
      }
   }

   internal class UndoMoveAction {
      PlayingCard _card;
      internal PlayingCard Card {
         get { return _card; }
      }

      internal UndoMoveAction(PlayingCard Card) {
         _card = Card;
      }
   }

   internal class CardFlipAction : UndoMoveAction {
      bool _faceVisibleAtStart;
      internal bool FaceVisibleAtStart {
         get { return _faceVisibleAtStart; }
      }

      internal CardFlipAction(PlayingCard Card) : base(Card) {
         _faceVisibleAtStart = !Card.IsFaceShowing;
      }
   }

   internal class CardRemoveAction : UndoMoveAction {
      PlayingStack _removedFromStack;
      internal PlayingStack RemovedFromStack {
         get { return _removedFromStack; }
      }

      internal CardRemoveAction(PlayingCard Card) : base(Card) {
         _removedFromStack = Card.Stack;
      }
   }

   internal class CardAddAction : UndoMoveAction {
      PlayingStack _addedToStack;
      internal PlayingStack AddedToStack {
         get { return _addedToStack; }
      }

      internal CardAddAction(PlayingCard Card) : base(Card) {
         _addedToStack = Card.Stack;
      }
   }

   internal class GameSpecificChangedAction : UndoMoveAction {
      String _gameSpecific;
      internal String GameSpecificInfo {
         get { return _gameSpecific; }
      }

      IRules _currentGame;
      internal IRules CurrentGame {
         get { return _currentGame; }
      }

      internal GameSpecificChangedAction(IRules CurrentGame, String GameSpecificInfo) : base(null) {
         _currentGame = CurrentGame;
         _gameSpecific = GameSpecificInfo;
      }
   }
}
